$| =1;
#####################################################################
# Spexsubs provides a number of functions which can be used either
# by the spex master, the spex slave or the spex standalone.
#####################################################################




################################
################################
# Libraries
################################
################################

use Sys::Hostname;
use Cwd;
use IO::Handle;
use File::Basename;
use Getopt::Std;
use IPC::Open2;
use IO::Socket;
use IO::Pipe;
use IO::Select;
use IO::File;
require "flush.pl";




################################
################################
# Global Data 
################################
################################

# Escape character surrounding variable names in a model
# template file
$pesc = "#!#";

# Some regular expressions
$exp_re = "[Ee][+-]?\\d+";
$real_re = "[+-]?\\d+(\\.\\d*)?($exp_re)?";
$name_re = "\\\$[_a-zA-Z]\\w*";

# Comment character
$cmt = '#';

# Stores task throughputs parsed from model output file
%xtask = ();

# Stores waiting time for open arrivals
%openwaiting = ();

# Stores entry throughputs parsed from model output file
%xentry = ();

# Stores entry service times parsed from model output file
%entrysvc = ();

# Stores entry utilizations parsed from model output file
%uentry = ();

# Stores task utilizations parsed from model output file
%utask = ();

# Stores entry service time variance parsed from model output file
%entryvar = ();

$test_results = 0;
%uproc = ();

# Stores throughput bounds parsed from model output file
%bounds = ();

# Stores convergence info of solver
%tests = ();

# Stores rendez-vous waiting times parsed from model output file
%rnvwaiting = ();

# Stores variance of rendez-vous waiting times parsed from model output file
%rnvwaitvar = ();

# Stores processor utilization of entries parsed from model output file
%upentry = ();

# Stores processor utilization of tasks parsed from model output file
%uptask = ();

# Stores processor waiting time of entriess parsed from model output file
%wpentry = ();




################################
################################
# Function: apply_solver_to_case
# - if agr preprocessing is required,
#   first call spexagr.  Then invoke
#   the solver 
################################
################################

sub apply_solver_to_case {
    my($newin,$cmd,@cmdout);
    my $agrspec = $_[4];
    if ($agrspec) {
	$cmd = "spexagr -s $_[2].agr -i $_[0].in -o $_[0].in; $_[1] -p -o $_[0].out $_[0].in";
    } else {
	$cmd = "$_[1] -p -o $_[0].out $_[0].in";
    }

    open(CMD,"$cmd"." |");
    @cmdout = <CMD>;
    close CMD;
    $rc = $? / 256;
    if ( $rc != 0 ) {
	$error_msg = "@cmdout";
	$error_rc = $rc;
    }
}




################################
################################
# Function: pipe_package
# - replace all new lines in a message
#   with <newline>.  Prepares a mutli-line message
#   so that it can be sent over a pipe/socket
#   with one write call
################################
################################

sub pipe_package {
    my($msg) = $_[0];
    $$msg =~ s/\n/<newline>/g;
}




################################
################################
# Function: pipe_unpackage
# - replace all <newline>'s stuffed in
#   by pipe_package with real new lines
################################
################################

sub pipe_unpackage {
    my($msg) = $_[0];
    $$msg =~ s/<newline>/\n/g;
}




################################
################################
# Function: print_agr_parm
# - Write out parameter names and values for agr reduction 
#   for a given case
################################
################################

sub print_agr_parm {
    my $paramsptr = $_[1];
    $agrfile = new IO::File $_[0], "w";
    if (!defined($agrfile)) {
	$error_msg = "Cannot write to $_[0]\n";
	$error_rc = -1;
	return;
    }
    while ( ($name,$value) = each %{$paramsptr} ) {    
	print $agrfile "$name $value \n";
    }
    close $agrfile;
}




################################
################################
# Function: collect_selected_results
# - Parse through the model output file (.p file)
#   retrieving the information requested
#   with the tag variables in the spex
#   model definition file.  
#   If you don't understand this, look at
#   the pformat file in this directory. 
#   It contains the .p file.  Check with Greg
#   to make sure it hasn't changed.
################################
################################

sub collect_selected_results {
    my ( $filename, $line, $phases, $state, $rdata, $current_task, $current_processor, $r, $id, @terms, $e1, $e2, $paramsptr, $rtagsptr );
    ($filename, $paramsptr, $rtagsptr) = @_;

    open( RESULT_FILE, "<$filename" ) || die "Unable to open model results file \"$filename\" -- $!";

    $line = <RESULT_FILE>;
    die "Result file is empty $filename -- $!" unless defined $line;

    while ( $line = <RESULT_FILE> )
    {
       if (($line !~ /^\s*#.*/) & ($line !~ /^\s*$/))
       {
          last;
       }
    }

    $line =~ /^V\s([yn])$/;
    die "Bad result file format -- unable to read validity flag\n" unless defined $line;
    if ( $1 eq "n" ) {
	print "WARNING: The model for case \"$filename\" did not converge.\n";
    }	

    $line = <RESULT_FILE>;
    die "Result file is empty $filename -- $!" unless defined $line;
    $line =~ /C\s($real_re)/;
    die "Bad result file format -- unable to read convergence\n" unless defined $line;
    $tests{'convergence'} = $1;

    $line = <RESULT_FILE>;
    die "Result file is empty $filename -- $!" unless defined $line;
    $line =~ /I\s(\d+)/;
    die "Bad result file format -- unable to read iterations\n" unless defined $line;
    $tests{'iterations'} = $1;

    $line = <RESULT_FILE>;
    die "Result file is empty $filename -- $!" unless defined $line;
    $line =~ /PP\s\d+/;
    die "Bad result file format - uanble to read number of processors\n" unless defined $line;

    $line = <RESULT_FILE>;
    die "Result file is empty $filename -- $!" unless defined $line;
    $line =~ /NP\s(\d+)/;
    die "Bad result file format -- unable to read number of phases\n" unless defined $line;
    $phases = $1;

    # Go To State: Initial
    $state = 0;
    while ( <RESULT_FILE> ) {

	next if /^\s*#.*/ || /^\s*$/;

	# First section is entry throughput bounds
	if ( /^B\s\d+\s*/ ) {
	    # Go To State: Reading Entry Throughput Bounds
	    $state = 1;
	    next;
	}

        # State: Reading Entry Throughput Bounds
	if ( $state == 1 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
  	        # Go To State: Initial
		$state = 0;
		next;

	    # Read in bounds
	    } elsif ( /^\s*(\w+\s*:\s*)?(\w+)\s+($real_re)\s*$/ ) {
		$bounds{$2} = $3;
		next;

	    } else {
		print "WARNING: Unexpected input in bounds section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# Next we have the waiting time for a rendezvous
	if ( /^W\s\d+\s*/ ) {
	    # Go To State: Reading Rendezvous Waiting Times
	    $state = 2;
	    next;
	}

	#State: Reading Rendezvous Waiting times
	if ( $state == 2 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
  	        # Go To State: Initial
		$state = 0;
		next;

	    } elsif (/^\s*\w+\s*:\s*-\s*1\s*$/) {
		next;
	    } elsif (/^\s*\w+\s*:\s*-\s*1\s*:\s*$/) {
		# Go To State: Reading Waiting Times for Task Activities		
		$state=22;
		next;
	    } elsif (/^\s*:\s*$/) {
		# Go To State: Reading Waiting Times for Task Activities		
		$state = 22;
		next;
	    } elsif ( /^\s*\w+\s*:\s*(\w+)\s+(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Waiting Times for Task Entries
		$e1 = $1;
		$e2 = $2;
		$rnvwaiting{"$1:$2"} = get_phase_list( $3 );
		$state = 21; 
		next;
	    } elsif ( /^\s*(\w+)\s+(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Waiting Times for Task Entries
		$e1 = $1;
		$e2 = $2;
		$state = 21; 
		$rnvwaiting{"$1:$2"} = get_phase_list( $3 );
		next;
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaiting{"$e1:$e2:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaiting{"$e1:$e2:99"} = get_phase_list( $1 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in rendezvous waiting section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Waiting Times for Task Entries
	if ( $state == 21 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		#Go To State: Reading Rendezvous Waiting times
		$state = 2;
		next;
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaiting{"$e1:$e2:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaiting{"$e1:$e2:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		$e1 = $1;
		$e2 = $2;
		$rnvwaiting{"$1:$2"} = get_phase_list( $3 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in rendezvous waiting section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Waiting Times for Task Activities
	if ($state == 22) {
	    if ( /^\s*-\s*1\s*$/ ) {
		#Go To State: Reading Rendezvous Waiting times
		$state = 2;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+-1\s*$/ ) {
		$rnvwaiting{"a:$a1:$e2:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+-1\s*$/ ) {
		$rnvwaiting{"a:$a1:$e2:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+(\w+)\s+($real_re)\s+-1\s*$/ ) {
		$a1 = $1;
		$e2 = $2;
		$rnvwaiting{"a:$1:$2"} = get_phase_list( $3 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in rendezvous waiting section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# Next we have the variances on the waiting times
	if ( /^VARW\s\d+\s*/ ) {
	    # Go To State: Reading Variance for Rendezvous Waiting TImes
	    $state = 30;
	    next;
	}

        # State: Reading Variance for Rendezvous Waiting TImes
	if ( $state == 30 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	 	# Go To State: Initial
		$state = 0;
		next;
	    } elsif (/^\s*\w+\s*:\s*-\s*1\s*$/) {
		next;
	    } elsif (/^\s*\w+\s*:\s*-\s*1\s*:\s*$/) {
		# Go To State: Reading Variance of Waiting Times for Task Activities
		$state=33;
		next;
	    } elsif (/^\s*:\s*$/) {
		# Go To State: Reading Variance of Waiting Times for Task Activities
		$state = 33;
		next;
	    } elsif ( /^\s*\w+\s*:\s*(\w+)\s+(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Variance of Waiting Times for Task Entries
		$e1 = $1;
		$e2 = $2;
		$rnvwaitvar{"$1:$2"} = get_phase_list( $3 );
		$state = 31; 
		next;
	    } elsif ( /^\s*(\w+)\s+(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Variance of Waiting Times for Task Entries
		$e1 = $1;
		$e2 = $2;
		$rnvwaitvar{"$1:$2"} = get_phase_list( $3 );
		$state = 31; 
		next;
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaitvar{"$e1:$e2:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaitvar{"$e1:$e2:99"} = get_phase_list( $1 );
		next;
	    } else {
	 	# Go To State: Initial
		print "WARNING: Unexpected input in rendezvous waiting section (state = $state)\n";
		print " => $_";
		$state = 0;  print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Variance of Waiting Times for Task Entries
	if ( $state == 31 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	        # Go To State: Reading Variance for Rendezvous Waiting TImes
		$state = 30;
		next;
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaitvar{"$e1:$e2:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$rnvwaitvar{"$e1:$e2:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		$e1 = $1;
		$e2 = $2;
		$rnvwaitvar{"$1:$2"} = get_phase_list( $3 );
		next;
	    } else {
	 	# Go To State: Initial
		print "WARNING: Unexpected input in rendezvous waiting section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Variance of Waiting Times for Task Activities
	if ( $state == 33 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	        # Go To State: Reading Variance for Rendezvous Waiting TImes
		$state = 30;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+-1\s*$/ ) {
		$rnvwaitvar{"a:$a1:$e2:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+-1\s*$/ ) {
		$rnvwaitvar{"a:$a1:$e2:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+(\w+)\s+($real_re)\s+-1\s*$/ ) {
		$a1 = $1;
		$e2 = $2;
		$rnvwaitvar{"a:$1:$2"} = get_phase_list( $3 );
		next;
	    } else {
	 	# Go To State: Initial
		print "WARNING: Unexpected input in rendezvous waiting section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# Next we have the service times
	if ( /^X\s\d+\s*/ ) {
	    # Go To State: Reading Service TImes
	    $state = 3;
	    next;
	}

	# State: Reading Service Times
	if ( $state == 3 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Initial
		$state = 0;
		next;
	    } elsif (/^\s*\w+\s*:\s*-\s*1\s*$/) {
		next;
	    } elsif (/^\s*\w+\s*:\s*-\s*1\s*:\s*$/) {
		# Go To State: Reading Activity Service Time
		$state=34;
		next;
	    } elsif (/^\s*:\s*$/) {
		# Go To State: Reading Activity Service Time
		$state = 34;
		next;
	    } elsif ( /^\s*\w+\s*:\s*(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Entry Service Time
		$e1 = $1;
		$entrysvc{$1} = get_phase_list( $2 );
		$state=32;
		next;
	    } elsif ( /^\s*(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Entry Service Time
		$e1 = $1;
		$entrysvc{$1} = get_phase_list( $2 );
		$state=32;
		next;
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$entrysvc{"$e1:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$entrysvc{"$e1:99"} = get_phase_list( $1 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in service time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Entry Service Time
	if ($state == 32) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Service Times
		$state = 3;
		next;	    
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$entrysvc{"$e1:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$entrysvc{"$e1:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		$e1 = $1;
		$entrysvc{"$1"} = get_phase_list( $2 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in service time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Activity Service Time
	if ($state == 34) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Service Times
		$state = 3;
		next;	    
	    } elsif ( /^\s*%\s95\s+($real_re)\s+-1\s*$/ ) {
		$entrysvc{"a:$a1:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+-1\s*$/ ) {
		$entrysvc{"a:$a1:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+($real_re)\s+-1\s*$/ ) {
		$a1 = $1;
		$entrysvc{"a:$1"} = get_phase_list( $2 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in service time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# Next we have the service time variances
	if ( /^VAR\s\d+\s*/ ) {
	    # Go To State: Reading Variance of Service Times
	    $state = 4;
	    next;
	}

        # State: Reading Variance of Service Times
	if ( $state == 4 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Initial
		$state = 0;
		next;
	    } elsif (/^\s*:\s*$/) {
		# Go To State: Reading Variance of Activity Service Time
		$state = 42;
		next;
	    } elsif ( /^\s*\w+\s*:\s*(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Variance of Entry Service Time
		$e1 = $1;
		$entryvar{$1} = get_phase_list( $2 );
		$state=41;
		next;
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$entryvar{"$e1:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$entryvar{"$e1:99"} = get_phase_list( $1 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in service time variance section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Variance of Entry Service Time
	if ( $state == 41 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	        # Go To State: Reading Variance of Service Times
		$state = 4;
		next;	    
	    } elsif ( /^\s*%\s95\s+(($real_re\s+)+)-1\s*$/ ) {
		$entryvar{"$e1:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+(($real_re\s+)+)-1\s*$/ ) {
		$entryvar{"$e1:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+(($real_re\s+)+)-1\s*$/ ) {
		$e1 = $1;
		$entryvar{"$1"} = get_phase_list( $2 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in service time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}	    

	# State: Reading Variance of Activity Service Time
	if ( $state == 42 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	        # Go To State: Reading Variance of Service Times
		$state = 4;
		next;	    
	    } elsif ( /^\s*%\s95\s+($real_re)\s+-1\s*$/ ) {
		$entryvar{"a:$a1:95"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+-1\s*$/ ) {
		$entryvar{"a:$a1:99"} = get_phase_list( $1 );
		next;
	    } elsif ( /^\s*(\w+)\s+($real_re)\s+-1\s*$/ ) {
		$a1 = $1;
		$entryvar{"a:$1"} = get_phase_list( $2 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in service time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}	    

	# Next we have the throughputs and utilizations.
	if ( /^FQ\s\d+\s*/ ) {
	    # Go To State: Reading Throughputs and Utilizations
	    $state = 5;
	    next;
	}

	# State: Reading Throughputs and Utilizations
	if ( $state == 5 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Initial
		$state = 0;
		next;
	    } elsif (/^\s*:\s*$/) {
		# Go To State: Reading Activity Throughput and Utilization
		$state = 52;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xtask{"$current_task:95"} = $1;
		$utask{"$current_task:95"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xtask{"$current_task:99"} = $1;
		$utask{"$current_task:99"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*(\w+)\s+(:|\d+)\s+(\w+)\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		# Go To State: Reading Entry Throughput and Utilization
		$current_task = $1;
		$e1 = $3;
		$xentry{$3} = $4;
		$uentry{$3} = get_phase_list( $7 );
		$xtask{$current_task} = $4;
		$utask{$current_task} = get_phase_list( $7 );
		$state = 6;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in throughtputs and utilization section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Activity Throughput and Utilization
	if ( $state == 52 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Throughputs and Utilizations
		$state = 5;
		next;
	    } elsif ( /^\s+(\w+)\s+($real_re)\s+(($real_re\s+)+)-1/ ) { # ignore restof line
		$a1 = $1;
		$xentry{"a:$1"} = $2;
		$uentry{"a:$1"} = get_phase_list( $5 );
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xentry{"a:$a1:95"} = $1;
		$uentry{"a:$a1:95"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xentry{"a:$a1:99"} = $1;
		$uentry{"a:$a1:99"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s+(\w+)\s+($real_re)\s+($real_re)\s+-1/ ) { # ignore restof line
		$a1 = $1;
		$xentry{"a:$1"} = $2;
		$uentry{"a:$1"} = get_phase_list( $3 );
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+($real_re)\s+-1\s+$real_re\s*$/ ) {
		$xentry{"a:$a1:95"} = $1;
		$uentry{"a:$a1:95"} = get_phase_list( $2 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+($real_re)\s+-1\s+$real_re\s*$/ ) {
		$xentry{"a:$a1:99"} = $1;
		$uentry{"a:$a1:99"} = get_phase_list( $4 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in throughtputs and utilization section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Entry Throughput and Utilization
	if ( $state == 6 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Throughputs and Utilizations
		$state = 5;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xentry{"$e1:95"} = $1;
		$uentry{"$e1:95"} = get_phase_list( $4 );
		$xtask{"$current_task:95"} = $1;
		$utask{"$current_task:95"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xentry{"$e1:99"} = $1;
		$uentry{"$e1:99"} = get_phase_list( $4 );
		$xtask{"$current_task:99"} = $1;
		$utask{"$current_task:99"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s+(\w+)\s+($real_re)\s+(($real_re\s+)+)-1/ ) { # ignore restof line
		# State: Reading More Entry Throughput and Utilization
		$e1 = $1;
		$xentry{$1} = $2;
		$uentry{$1} = get_phase_list( $5 );
		$state = 7;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in throughtputs and utilization section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading More Entry Throughput and Utilization
	if ( $state == 7 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Totals of Task Throughput and Utilization
		$state = 8;
		next;
	    } elsif ( /^\s+(\w+)\s+($real_re)\s+(($real_re\s+)+)-1/ ) { # ignore restof line
		$e1 = $1;
		$xentry{$1} = $2;
		$uentry{$1} = get_phase_list( $5 );
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xentry{"$e1:95"} = $1;
		$uentry{"$e1:95"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s+$real_re\s*$/ ) {
		$xentry{"$e1:99"} = $1;
		$uentry{"$e1:99"} = get_phase_list( $4 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in throughtputs and utilization section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Totals of Task Throughput and Utilization
	if ( $state == 8 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Initial
		$state = 0;
		next;
	    } elsif ( /^\s+($real_re)\s+(($real_re\s+)+)-1/ ) { # ignore restof line
		$xtask{$current_task} = $1;
		$utask{$current_task} = get_phase_list( $4 );
		$state = 5;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in throughtputs and utilization section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# Open arrival section
	if ( /^R\s\d+\s*$/ ) {
	    # Go To State: Reading Open Arrival Waiting Times
	    $state = 40;
	    next;
	}

	# State: Reading Open Arrival Waiting Times
	if ( $state == 40 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Initial
		$state = 0;
		next;
	    } elsif ( /^\s*(\w+\s*(:|\s)?\s*)?(\w+)\s+$real_re\s+($real_re)\s*$/ ) {
		$e1 = $3;
		$openwaiting{$3} = $6;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s*$/ ) {
		$openwaiting{"$e1:95"} = $1;
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s*$/ ) {
		$openwaiting{"$e1:99"} = $1;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in open arrival waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# Finally we have the processor utilization and waiting time information

	if ( /^P\s(\w+)\s\d+\s*$/ ) {
	    # Go To State: Reading Processor Information
	    $current_processor = $1;
	    $uproc{$current_processor} = 0;
	    $state = 10;
	    next;
	}

        # State: Reading Total and Next Processor Information
	if ( $state == 9 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Initial
		$state = 0;
		next;
	    } elsif ( /^\s*($real_re)\s*$/ ) {
		# Go To State: Reading Processor Totals
		$state = 14;
		next;
	    } elsif ( /^P\s(\w+)\s\d+\s*$/ ) {
		# Go To State: Reading Processor Information
		$current_processor = $1;
		$uproc{$current_processor} = 0;
		$state = 10;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

        # State: Reading Processor Information
	if ( $state == 10 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	        # Go To State: Reading Total and Next Processor Information
		$state = 9;
		next;
	    } elsif (/^\s*:\s*$/) {
		# Go To State: Reading Activity Processor Information
		$state = 15;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s*$/ ) {
		$uptask{"$current_task:95"} = $1;
		$state = 10;
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s*$/ ) {
		$uptask{"$current_task:99"} = $1;
		$state = 10;
		next;
	    } elsif ( /^\s*(\w+)\s+\d+\s+\d+\s+\d+\s+(\w+)\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Entry Processor Information
		$current_task = $1;
		$e1 = $2;
		$upentry{$2} = $3;
		$wpentry{$2} = get_phase_list( $6 );
		$uptask{$current_task} = $3;
		$state = 11;
		next;
	    } elsif ( /^\s*(\w+)\s+\d+\s+\d+\s+(\w+)\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading Entry Processor Information
		$current_task = $1;
		$e1 = $2;
		$upentry{$2} = $3;
		$wpentry{$2} = get_phase_list( $6 );
		$uptask{$current_task} = $3;
		$state = 11;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Entry Processor Information
	if ( $state == 11 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Processor Information
		$uproc{$current_processor} += $uptask{$current_task};
		$state = 10;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$upentry{"$e1:95"} = $1;
		$wpentry{"$e1:95"} = get_phase_list( $4 );
		$uptask{"$current_task:95"} = $1;
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$upentry{"$e1:99"} = $1;
		$wpentry{"$e1:99"} = get_phase_list( $4 );
		$uptask{"$current_task:99"} = $1;
		next;
	    } elsif ( /^\s*(\w+)\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		# Go To State: Reading More Entry Processor Information
		$e1 = $1;
		$upentry{$1} = $2;
		$wpentry{$1} = get_phase_list( $5 );
		$state = 12;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading More Entry Processor Information
	if ( $state == 12 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		# Go To State: Reading Processor Totals per Task
		$state = 13;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$upentry{"$e1:95"} = $1;
		$wpentry{"$e1:95"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$upentry{"$e1:99"} = $1;
		$wpentry{"$e1:99"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*(\w+)\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$e1 = $1;
		$upentry{$1} = $2;
		$wpentry{$1} = get_phase_list( $5 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Processor Totals per Task
	if ( $state == 13 ) {
	    if ( /^\s*($real_re)\s*$/ ) {
		# Go To State: Reading Processor Information
		$uptask{$current_task} = $1;
		#$uproc{$current_processor} += $1;
		$state = 10;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Processor Totals
	if ( $state == 14 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
	        # Go To State: Reading Total and Next Processor Information
		$state = 9;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s*$/ ) {
		$uproc{"$current_processor:95"} = $1;
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s*$/ ) {
		$uproc{"$current_processor:99"} = $1;
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}

	# State: Reading Activity Processor Information
	if ( $state == 15 ) {
	    if ( /^\s*-\s*1\s*$/ ) {
		$state = 13;
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$upentry{"a:$a1:95"} = $1;
		$wpentry{"a:$a1:95"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$upentry{"a:$a1:99"} = $1;
		$wpentry{"a:$a1:99"} = get_phase_list( $4 );
		next;
	    } elsif ( /^\s*(\w+)\s+($real_re)\s+(($real_re\s+)+)-1\s*$/ ) {
		$a1 = $1;
		$upentry{"a:$1"} = $2;
		$wpentry{"a:$1"} = get_phase_list( $5 );
		next;
	    } elsif ( /^\s*%\s95\s+($real_re)\s+($real_re)\s+-1\s*$/ ) {
		$upentry{"a:$a1:95"} = $1;
		$wpentry{"a:$a1:95"} = get_phase_list( $2 );
		next;
	    } elsif ( /^\s*%\s99\s+($real_re)\s+($real_re)\s+-1\s*$/ ) {
		$upentry{"a:$a1:99"} = $1;
		$wpentry{"a:$a1:99"} = get_phase_list( $2 );
		next;
	    } elsif ( /^\s*(\w+)\s+($real_re)\s+($real_re)\s+-1\s*$/ ) {
		$a1 = $1;
		$upentry{"a:$1"} = $2;
		$wpentry{"a:$1"} = get_phase_list( $3 );
		next;
	    } else {
		# Go To State: Initial
		print "WARNING: Unexpected input in processor utilization and waiting time section (state = $state)\n";
		print " => $_";
		$state = 0; print "Continuing... \n";
		next;
		exit 1;
	    }
	}
    }

    close RESULT_FILE;

    # print out contents of result hash tables if debugging turned on
    if ( $test_results ) {
	print "Case data collected:\n";
	print "  Entry throughput bounds:\n";
	print_hash( \%bounds );
	print "  Open arrival waiting time:\n";
	print_hash( \%openwaiting );
	print "  Rnv waiting time:\n";
	print_array_hash( \%rnvwaiting );
	print "  Entry service time:\n";
	print_array_hash( \%entrysvc );
	print " Entry variance:\n";
	print_array_hash( \%entryvar );
	print "  Entry throughput:\n";
	print_hash( \%xentry );
	print "  Task throughput:\n";
	print_hash( \%xtask );
	print "  Entry utilization:\n";
	print_array_hash( \%uentry );
	print "  Task utilization:\n";
	print_array_hash( \%utask );
	print "  Processor utilization by entry:\n";
	print_hash( \%upentry );
	print "  Processor utilization by task:\n";
	print_hash( \%uptask );
	print "  Processor utilization:\n";
	print_hash( \%uproc );
	print "  Processor waiting time by entry:\n";
	print_array_hash( \%wpentry );
	print "\n\n";
    }

    # for each tag variable in rtags, read in the variable name
    # and the type of information requested.  Extract this information
    # from the appropriate result hashtable.  If you recall from the spex
    # main program, the format of an rtags row is:
    # rtag variable name|type of info id | processor/task/entry name | type specific info...
    my @results = ();
    foreach $r (@{$rtagsptr}) {
	@terms = ($r =~ /(\w+)/g);
	$var = '$'.shift( @terms ); #'
  	$id = $terms[0];
      SWITCH: {

	  # Throughput bounds
	  if ( $id eq "b" ) {
	      $paramsptr->{$var} = $bounds{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of throughput bounds
	  if ( $id eq "bc" ) {
	      $paramsptr->{$var} = $bounds{"$terms[1]:$terms[2]"};
	      last SWITCH;
	  }

	  # Waiting time for open arrivals
	  if ( $id eq "wa" ) {
	      $paramsptr->{$var} = $openwaiting{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of waiting time for open arrivals
	  if ( $id eq "wac" ) {
	      $paramsptr->{$var} = $openwaiting{"$terms[1]:$terms[2]"};
	      last SWITCH;
	  }

	  # Waiting time for entry to entry rendez-vous
	  if ( $id eq "wr" ) {
	      $paramsptr->{$var} = $rnvwaiting{"$terms[1]:$terms[2]"}[$terms[3]];
	      last SWITCH;
	  }

	  # Confidence interval of waiting time for entry to entry rendez-vous
	  if ( $id eq "wrc" ) {
	      $paramsptr->{$var} = $rnvwaiting{"$terms[1]:$terms[2]:$terms[4]"}[$terms[3]];
	      last SWITCH;
	  }

	  # Variance of waiting time for entry to entry rendez-vous
	  if ( $id eq "vr" ) {
	      $paramsptr->{$var} = $rnvwaitvar{"$terms[1]:$terms[2]"}[$terms[3]];
	      last SWITCH;
	  }

	  # Confidence interval of variance of waiting time for entry to entry rendez-vous
	  if ( $id eq "vrc" ) {
	      $paramsptr->{$var} = $rnvwaitvar{"$terms[1]:$terms[2]:$terms[4]"}[$terms[3]];
	      last SWITCH;
	  }

          # Entry service time
	  if ( $id eq "s" ) {
	      $paramsptr->{$var} = ${$entrysvc{$terms[1]}}[$terms[2]];
	      last SWITCH;
          }

          # Confidence interval of entry service time
	  if ( $id eq "sc" ) {
	      $paramsptr->{$var} = ${$entrysvc{"$terms[1]:$terms[3]"}}[$terms[2]];
	      last SWITCH;
          }

	  # Variance of entry service time
	  if ( $id eq "sv" ) {
	     $paramsptr->{$var} = ${$entryvar{$terms[1]}}[$terms[2]];
	      last SWITCH;
	  }

          # Confidence interval of variance entry service time
	  if ( $id eq "svc" ) {
	     $paramsptr->{$var} = ${$entryvar{"$terms[1]:$terms[3]"}}[$terms[2]];
	      last SWITCH;
	  }

	  # Entry throughput
	  if ( $id eq "xe" ) {
	      $paramsptr->{$var} = $xentry{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of entry throughput
	  if ( $id eq "xec" ) {
	      $paramsptr->{$var} = $xentry{"$terms[1]:$terms[2]"};
	      last SWITCH;
          }

	  # Task throughput
	  if ( $id eq "xt" ) {
	      $paramsptr->{$var} = $xtask{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of task throughput
	  if ( $id eq "xtc" ) {
	      $paramsptr->{$var} = $xtask{"$terms[1]:$terms[2]"};
	      last SWITCH;
	  }

	  # Entry utilization
	  if ( $id eq "ue" ) {
	      $paramsptr->{$var} = ${$uentry{$terms[1]}}[$terms[2]];
	      last SWITCH;
	  }

	  # Confidence interval of entry utilization
	  if ( $id eq "uec" ) {
	      $paramsptr->{$var} = ${$uentry{"$terms[1]:$terms[3]"}}[$terms[2]];
	      last SWITCH;
	  }

	  # Task utilization
	  if ( $id eq "ut" ) {
	      $paramsptr->{$var} = ${$utask{$terms[1]}}[$terms[2]];
	      last SWITCH;
          }

	  # Confidence interval of task utilization
	  if ( $id eq "utc" ) {
	      $paramsptr->{$var} = ${$utask{"$terms[1]:$terms[3]"}}[$terms[2]];
	      last SWITCH;
	  }

	  # Entry processor utilization
	  if ( $id eq "upe" ) {
	      $paramsptr->{$var} = $upentry{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of entry processor utilization
	  if ( $id eq "upec" ) {
	      $paramsptr->{$var} = $upentry{"$terms[1]:$terms[2]"};
	      last SWITCH;
	  }

	  # Task processor utilization
	  if ( $id eq "upt" ) {
	      $paramsptr->{$var} = $uptask{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of task processor utilization
	  if ( $id eq "uptc" ) {
	      $paramsptr->{$var} = $uptask{"$terms[1]:$terms[2]"};
	      last SWITCH;
	  }

	  # Overall processor utilization
	  if ( $id eq "up" ) {
	      $paramsptr->{$var} = $uproc{$terms[1]};
	      last SWITCH;
	  }

	  # Confidence interval of overall processor utilization
	  if ( $id eq "upc" ) {
	      $paramsptr->{$var} = $uproc{"$terms[1]:$terms[2]"};
	      last SWITCH;
	  }

	  # Processor waiting time for entry
	  if ( $id eq "wpe" ) {
	      $paramsptr->{$var} = ${$wpentry{$terms[1]}}[$terms[2]];
	      last SWITCH;
          }

	  # Confidence interval of processor waiting time for entry
	  if ( $id eq "wpec" ) {
	      $paramsptr->{$var} = ${$wpentry{"$terms[1]:$terms[3]"}}[$terms[2]];
	      last SWITCH;
	  }
	  print "ERROR: Unrecognized result id: \"$id\"\n";
	  print " => Result specification: $r\n";
	  exit 1;
       }
    }

    return @results;
}




################################
################################
# Function: evaluate_param_exprs
# - Given a set of parameter values,
#   calculate the value for those
#   parameters which are expressions
#   of others.  
################################
################################

sub evaluate_param_exprs {
    my( $i, %exprs, $name, $value );
    my ($paramsptr, $param_lists_namesptr, $param_lists_valuesptr, $paraminfoptr, $componentsptr) = @_;

    # First we add all the parameter list parameters
    for ( $i = 0; $i <= $#{$componentsptr}; $i++ ) {
	$paramsptr->{$param_lists_namesptr->[$i]} = ${$param_lists_valuesptr->[$i]}[$componentsptr->[$i]];
    }

    # Copy the parameter info over -- we'll need it to be intact for the
    # subsequent cases.
    while ( ($name,$value) = each %{$paraminfoptr} ) {
	$exprs{$name} = $value;
    }

    # Evaluate the expressions as much as possible
    perform_evaluation( \%exprs, $paramsptr );

    # Double-check the results
    while ( ($name,$value) = each %{$paramsptr} ) {
	if ( !defined $value ) {
	    print "ERROR: Bad value for variable \"$name\"\n";
	    exit 1;
	}
    }
}




################################
################################
# Function: evaluate_result_exprs
# - Given a set of tag variable and parameter values,
#   calculate the value for the result variables
#   result values are stored in paramsptr array
################################
################################

sub evaluate_result_exprs {
    my( %exprs, $name, $value );
    my($paramsptr, $resultinfoptr) = @_;

    # Copy the result info over -- we'll need it to be intact for the
    # subsequent cases.
    while ( ($name,$value) = each %{$resultinfoptr} ) {
	$exprs{$name} = $value;
    }

    # Evaluate the expressions as much as possible
    perform_evaluation( \%exprs, $paramsptr );

    # Double-check the results
    while ( ($name,$value) = each %{$paramsptr} ) {
	if ( !defined $value ) {
	    print "ERROR: Bad value for variable \"$name\"\n";
	    exit 1;
	}
    }
}




################################
################################
# Function: evaluate_result_exprs
# - Given a set of result, tag variable and parameter values,
#   calculate the value for the convergence variables
#   Convergence values are stored in paramsptr array
################################
################################

sub evaluate_converge_exprs {
    my( %exprs, $name, $value );
    my($paramsptr, $convergeinfoptr) = @_;
    # Copy the result info over -- we'll need it to be intact for the
    # subsequent cases.
    while ( ($name,$value) = each %{$convergeinfoptr} ) {
	$exprs{$name} = $value;
    }

    # Evaluate the expressions as much as possible
    perform_evaluation( \%exprs, $paramsptr );

    # Double-check the results
    while ( ($name,$value) = each %{$paramsptr} ) {
	if ( !defined $value ) {
	    print "ERROR: Bad value for variable \"$name\"\n";
	    exit 1;
	}
    }
}




################################
################################
# Function: perform_evaluation
# - Since variable values can depend on expressions
#   of other variables, and those variables in turn
#   can depend on expressions of others, we will need
#   to resolve these expressions recursively.  Each invokation
#   iterates through the hash table of variables and resolves as 
#   many variable expressions as possible.  If some are unresolved,
#   perform_evaluation is invoked recursively
################################
################################

sub perform_evaluation {
    my( $name, $value, %nexprs, $count, $sum );
    my ($exprsptr, $paramsptr) = @_;
    $count = 0;
    $sum = 0;

    # For each variable/expression pair in exprsptr, attempt to
    # resolve the epxression
    while ( ($name,$value) = each %{$exprsptr} ) {
	my( $rstr, $n );

	# Resolve as many variables as possible in this expression
	($rstr,$n) = elim_vars( $value, $paramsptr );

	# If all variables in this expression have been resolved,
	# evaluate the expression and record the fact that a new
	# expression has been resolved
	if ( $n == 0 ) {
	    $paramsptr->{$name} = eval( $rstr );
	    if ( $@ ) {
		print "ERROR: Unable to evaluate expression \"$rstr\": $@\n";
		exit 1;
	    }
	    $count += 1;

	# Otherwise, store the semi-resolved expression in nexprs and record
	# that an expression will need to be resolved on the next invokation
	} else {
            $nexprs{$name} = $rstr;
	    $sum += $n;
        }
    }

    # If some of the variable expressions have been resolved but there remain
    # others which need further resolving, recursively call perform evaluation
    if ( $count > 0 and $sum > 0 ) {
	perform_evaluation( \%nexprs, $paramsptr );

    # Otherwise, if some of the expressions still need resolving but there has
    # been no resolving of any variable expressions on this invocation, then there
    # is an error
    } else {
	if ( $sum > 0 ) {
	    print "ERROR: There were unresolved expressions:\n";
	    while ( ($name,$value) = each %nexprs ) {
		print "   $name = $value\n";
	    }
	    exit 1;
	}
    }
}




################################
################################
# Function: elim_vars
# - for a given expression, substitute
#   the variable names with their values
#   stored in the paramsptr hash table
################################
################################

sub elim_vars {
    my ( @vars, @vlist, $var, $n, $val, $var_re );
    my ($str, $paramsptr) = @_;

    # Get a list of variables
    @vars = ($_[0] =~ /($name_re)/g);

    # Make the list so that each element is unique
    @vlist = ();
    foreach $var (@vars) {
	$found = 0;
	foreach $var2 (@vlist) {
	    if ( $var eq $var2 ) {
		$found = 1;
		last;
	    }
	}
	if ( !$found ) {
	    push( @vlist, $var );
	}
    }

    # Eliminate the variables keeping track of how many are left to
    # do.
    $n = $#vlist + 1;
    $str = $_[0];
    foreach $var (@vlist) {
	$val = $paramsptr->{$var};
	if ( defined( $val ) ) {
	    $str =~ s/\Q$var\E\b/$val/g;
	    $n -= 1;
	}
    }

    # Return the new expression string and the number of variables which still
    # need resolving.
    return ($str, $n);
}




################################
################################
# Function: fill_template
# - a model template is a parameterized model
#   file with tag variables, result section
#   and convergence section.  This function
#   replaces the parameters with their values
#   from the paramsptr hashtable.  When a template
#   has been filled, it is ready to send to the
#   solver
################################
################################

sub fill_template {
    my ($outfile, $templateptr, $paramsptr) = @_;

    $ofile = new IO::File  "$outfile", "w";
    if (! (defined $ofile)) {
	$error_msg =  "Could not open $outfile \n";
	$error_rc = -1;
	return;
    }
    my $line;

    # for each line in the template file, iterate through
    # the parameters in paramsptr and replace all parameters
    # with their values
    foreach $line (@{$templateptr}) {
	my $newline = $line;
	while (($name,$val) = (each %{$paramsptr} )) {
	    $var = $pesc.$name.$pesc;
	    $newline =~ s/\Q$var\E/$val/g;
	}
	print $ofile $newline;
    }

    $ofile->close();

}




################################
################################
# Function: get_phase_list
# - parse list of numeric values 
#   in a phase list
################################
################################

sub get_phase_list {
    my $plist = [];
    my (@pl, $i, $sum);

    @pl = ($_[0] =~ /(\S+)/g);
    $sum = 0;
    for ( $i = 0; $i <= $#pl; $i++ ) {
	if ( $pl[$i] =~ /($real_re)\s*/ ) {
	    push( @$plist, $1 );
	    $sum += $1;
	} else {
	    print "ERROR: Phase list must contain numeric data\n";
	    print " => $_[0]";
	}
    }
    unshift( @$plist, $sum );
    return $plist;
}




################################
################################
# Function: solve
# - for a given template and set of
#   parameter values, fill in the
#   template and solve the model.
#   If convergence is specified, the
#   model will be solved repeatedly
#   until the convergence variables
#   converge.  If activity graph reduction
#   is required, spex agr will be called 
#   to preprocess the model file before sending
#   it to the solver
################################
################################

sub solve {
    my($casename, $paramsptr, $controlsptr, $templateptr, $agrspec,
       $cspec, $resultinfoptr, $rtagsptr, $convergeinfoptr, $opt_n) = @_;

    %prevparams = %{$paramsptr};
    my $keep_converging = 1;
    my $itercnt = 0;

    # If converging, prepare convergence report file
    if ($cspec && !($opt_n)) {
	$cfile = new ConvFile($casename, $controlsptr, $convergeinfoptr, '#');
    }

    # Solve once.  If convergence required, keep solving until convergence occurs
    # or convergence_iterations reached
    while ($keep_converging) {
	$keep_converging = 0;

	# If converging, make sure we haven't exceeded iteration limit
	if ($cspec && !($opt_n)) {
	    $itercnt++;
	    if ($itercnt > $controlsptr->{'$convergence_iters'}) {
		$cfile->print("WARNING: Could not converge for case $casename\n");
		last;
	    }
	}

	# Flesh out template
	fill_template("$casename.in", $templateptr, $paramsptr);
	if ($error_rc !=0) {
	    last;
	}	    

	# If agr required, print the substitution file for agr
	if ($agrspec) {
	    print_agr_parm("$casename.agr", $paramsptr);
	    if ($error_rc !=0) {
		last;
	    }	    
	}

	# Solve the low-level model with lqns
	if (! $opt_n) {
	    apply_solver_to_case( "$casename", $controlsptr->{'$solver'} , $agrspec ); 
	    if ($error_rc !=0) {
		last;
	    }

	    # Check for convergence if required
	    if ($cspec) {
		# Collect results
		collect_selected_results("$casename.p", $paramsptr, $rtagsptr);
		evaluate_result_exprs($paramsptr, $resultinfoptr);

		# Convergence?
		evaluate_converge_exprs($paramsptr, $convergeinfoptr);
		my $a = $controlsptr->{'$convergence_under_relax'};
		foreach $p (keys %{$convergeinfoptr}) {
		    my $prev = $prevparams{$p};
		    my $cur = $paramsptr->{$p};
		    my $relaxed = (($a * $cur) + ((1 - $a) * $prev));
		    my $convergence = abs($cur - $prev) / $prev;
		    $cfile->append($prev,$cur,$relaxed,$convergence);
		    $prevparams{$p} = $relaxed;
		    if ($convergence > $controlsptr->{'$convergence_limit'}) {
			$keep_converging = 1;
		    }
		}
		$cfile->nextline();
		if ($keep_converging) {
		    %{$paramsptr} = %prevparams;
		}
	    }
	}
    }
    if ($cspec && !($opt_n)) {
	$cfile->close();
    }
}




################################
################################
# Function: read_in_template
# - open template file (which
#   was produced by the spex 
#   master) and read the contents
#   into the template array
################################
################################

sub read_in_template {
    my $tempfile = $_[0];
    $tfile = new IO::File "$templfile", "r";
    if (! (defined $tfile)) {
	$error_msg = "Template file $templfile not found \n";
	$error_rc = -1;
	return;
    }

# if (defined(@array) is deprecated in newer Perls
#    if (defined(@template = <$tfile>)) {  }
    if (@template = <$tfile>) {  }
    else { 
	$error_msg = "Could not read template file \n";
	$error_rc = -1;
    }
    $tfile->close();
    return @template;
}




################################
################################
# Function: tpf
# - Truncated power function.
#   Don't ask me why I put this
#   function in this file
################################
################################

sub tpf {
    return floor($_[0],0);
}




################################
################################
# Function: floor
# - Return the lower of the two arguments
#   Don't ask me why I put this
#   function in this file
################################
################################

sub floor {
    my ($val, $floor) = @_;
    if ($val > $floor) {
	return $val;
    } else {
	return $floor;
    }
}




################################
################################
# Function: floor
# - Ethernet calcuation stuff
#   Don't ask me why I put this
#   function in this file
################################
################################

sub calc_prob_one_trans {
    my $sim = $_[0];

    if ($sim <= 1) {
	return 1;
    } else {
	return ( 1 - ( 1 / $sim)) ** ($sim - 1);
    }
}




################################
################################
# Class: Convfile
# - I'm just experimenting here,
#   using an object-oriented approach
#   to write the convergence
#   report file
################################
################################

package ConvFile; @ISA = qw(IO::File);

################################
################################
# Class: Convfile Method: new
# - Create a new convergence report file
################################
################################

sub new {
    my $class = shift;
    my $casename = shift;
    my $controlsptr = shift;
    my $convergeinfoptr = shift;
    my $cmt = shift;

    # Open the convergence report file for writing
    my $self = new IO::File "$casename.conv", "w";
    $self->autoflush();

    # Print a comment line
    for ( my $i = 0; $i < 70; $i++ ) {
	$self->print("$cmt");
    }

    # Print out the convergence control info
    $self->print("\n");
    $self->print("$cmt CONVERGENCE CONTROL:\n");
    while ( ($name,$value) = (each %{$controlsptr})) {
	if ($name =~ /convergence/) {
	    $self->print("$cmt $name = $value \n");
	}
    }

    # Print out the expressions for the convergence variables
    $self->print("$cmt \n");
    $self->print("$cmt CONVERGENCE PARAMETERS:\n");
    while ( ($name,$value) = (each %{$convergeinfoptr})) {
	$self->print("$cmt $name = $value \n");
    }


    # Print out the convergence variable names as column headers
    for ( my $i = 0; $i < 70; $i++ ) {
	$self->print("$cmt");
    }
    $self->print("\n");
    $self->print("$cmt ");
    foreach $name (keys %{$convergeinfoptr}) {
	$self->printf("%-52s",$name);
    }
    $self->print("\n");

    # Print out column sub-headers for each convergence variable
    $self->print("$cmt ");
    foreach $name (keys %{$convergeinfoptr}) {
	$self->printf("%-13s%-13s%-13s%-13s","PREVIOUS","CURRENT","NEXT","CONVERGE");
    }
    $self->print("\n");

    # Return reference to a new convergence report file object
    bless $self;
}

################################
################################
# Class: Convfile Method: append
# - Write out the previous, and current,
#   next value of a convergence variable
#   along with the convergence stopping
#   condition
################################
################################

sub append {
    my($self,$prev,$cur,$next,$converge) = @_;
    $self->printf("%1.6e %1.6e %1.6e %1.6e",$prev,$cur,$next,$converge);
}

################################
################################
# Class: Convfile Method: nextline
# - Print out a new line
################################
################################

sub nextline {
    my $self = shift;
    $self->print("\n");
}

package main;

return TRUE;
